import argparse
import re


class FlagAction(argparse.Action):
    # From http://bugs.python.org/issue8538

    def __init__(self, option_strings, dest, default=None,
                 required=False, help=None, metavar=None,
                 positive_prefixes=['--'], negative_prefixes=['--no-']):
        self.positive_strings = set()
        self.negative_strings = set()
        for string in option_strings:
            assert re.match(r'--[A-z]+', string)
            suffix = string[2:]
            for positive_prefix in positive_prefixes:
                self.positive_strings.add(positive_prefix + suffix)
            for negative_prefix in negative_prefixes:
                self.negative_strings.add(negative_prefix + suffix)
        strings = list(self.positive_strings | self.negative_strings)
        super(FlagAction, self).__init__(
            option_strings=strings, dest=dest,
            nargs=0, const=None, default=default, type=bool, choices=None,
            required=required, help=help, metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.positive_strings:
            setattr(namespace, self.dest, True)
        else:
            setattr(namespace, self.dest, False)



from __future__ import print_function

import eventlet
eventlet.monkey_patch()  # noqa (code before rest of imports)

import errno
import inspect
import logging
import logging.config
import os
import re
import signal
import sys

import six
from eventlet import backdoor

from nameko import config
from nameko.exceptions import CommandError
from nameko.extensions import ENTRYPOINT_EXTENSIONS_ATTR
from nameko.runners import ServiceRunner


logger = logging.getLogger(__name__)

MISSING_MODULE_TEMPLATE = "^No module named '?{}'?$"


def is_type(obj):
    return isinstance(obj, six.class_types)


def is_entrypoint(method):
    return hasattr(method, ENTRYPOINT_EXTENSIONS_ATTR)


def import_service(module_name):
    parts = module_name.split(":", 1)
    if len(parts) == 1:
        module_name, obj = module_name, None
    else:
        module_name, obj = parts[0], parts[1]

    try:
        __import__(module_name)
    except ImportError as exc:
        if module_name.endswith(".py") and os.path.exists(module_name):
            raise CommandError(
                "Failed to find service, did you mean '{}'?".format(
                    module_name[:-3].replace('/', '.')
                )
            )

        missing_module_re = MISSING_MODULE_TEMPLATE.format(module_name)
        # is there a better way to do this?

        if re.match(missing_module_re, str(exc)):
            raise CommandError(exc)

        # found module, but importing it raised an import error elsewhere
        # let this bubble (resulting in a full stacktrace being printed)
        raise

    module = sys.modules[module_name]

    if obj is None:
        found_services = []
        # find top-level objects with entrypoints
        for _, potential_service in inspect.getmembers(module, is_type):
            if inspect.getmembers(potential_service, is_entrypoint):
                found_services.append(potential_service)

        if not found_services:
            raise CommandError(
                "Failed to find anything that looks like a service in module "
                "{!r}".format(module_name)
            )

    else:
        try:
            service_cls = getattr(module, obj)
        except AttributeError:
            raise CommandError(
                "Failed to find service class {!r} in module {!r}".format(
                    obj, module_name)
            )

        if not isinstance(service_cls, type):
            raise CommandError("Service must be a class.")

        found_services = [service_cls]

    return found_services


def setup_backdoor(runner, port):
    def _bad_call():
        raise RuntimeError(
            'This would kill your service, not close the backdoor. To exit, '
            'use ctrl-c.')
    socket = eventlet.listen(('localhost', port))
    # work around https://github.com/celery/kombu/issues/838
    socket.settimeout(None)
    gt = eventlet.spawn(
        backdoor.backdoor_server,
        socket,
        locals={
            'runner': runner,
            'quit': _bad_call,
            'exit': _bad_call,
        })
    return socket, gt


def run(services, backdoor_port=None):
    service_runner = ServiceRunner()
    for service_cls in services:
        service_runner.add_service(service_cls)

    def shutdown(signum, frame):
        # signal handlers are run by the MAINLOOP and cannot use eventlet
        # primitives, so we have to call `stop` in a greenlet
        eventlet.spawn_n(service_runner.stop)

    signal.signal(signal.SIGTERM, shutdown)

    if backdoor_port is not None:
        setup_backdoor(service_runner, backdoor_port)

    service_runner.start()

    # if the signal handler fires while eventlet is waiting on a socket,
    # the __main__ greenlet gets an OSError(4) "Interrupted system call".
    # This is a side-effect of the eventlet hub mechanism. To protect nameko
    # from seeing the exception, we wrap the runner.wait call in a greenlet
    # spawned here, so that we can catch (and silence) the exception.
    runnlet = eventlet.spawn(service_runner.wait)

    while True:
        try:
            runnlet.wait()
        except OSError as exc:
            if exc.errno == errno.EINTR:
                # this is the OSError(4) caused by the signalhandler.
                # ignore and go back to waiting on the runner
                continue
            raise
        except KeyboardInterrupt:
            print()  # looks nicer with the ^C e.g. bash prints in the terminal
            try:
                service_runner.stop()
            except KeyboardInterrupt:
                print()  # as above
                service_runner.kill()
        else:
            # runner.wait completed
            break


def main(args):
    if '.' not in sys.path:
        sys.path.insert(0, '.')

    if "LOGGING" in config:
        logging.config.dictConfig(config['LOGGING'])
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')

    services = []
    for path in args.services:
        services.extend(
            import_service(path)
        )

    run(services, backdoor_port=args.backdoor_port)



"""Commands are defined in here, with imports inline, to avoid triggering
imports from other subcommands (e.g. `run` will cause an eventlet monkey-patch
which we don't want for `shell`)

"""
from .actions import FlagAction


class Command(object):
    name = None

    @staticmethod
    def init_parser(parser):
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def main(args):
        # import inline to avoid triggering imports from other subcommands
        raise NotImplementedError  # pragma: no cover


class Backdoor(Command):
    """Connect to a nameko backdoor.

    If a backdoor is running this will connect to a remote shell.  The
    runner is generally available as `runner`.
    """

    name = 'backdoor'

    @staticmethod
    def init_parser(parser):
        parser.add_argument(
            'target', metavar='[host:]port',
            help="(host and) port to connect to",
        )
        parser.add_argument(
            '--rlwrap', dest='rlwrap', action=FlagAction,
            help='Use rlwrap')
        parser.set_defaults(feature=True)
        return parser

    @staticmethod
    def main(args):
        from .backdoor import main
        main(args)


class ShowConfig(Command):
    """ Output as YAML string the configuration that would be passed to a
    service.

    Useful for viewing config files that load values from environement
    variables.
    """

    name = 'show-config'

    @staticmethod
    def init_parser(parser):
        return parser

    @staticmethod
    def main(args):
        from .show_config import main
        main(args)


class Run(Command):
    """Run nameko services.  Given a python path to a module containing one or
    more nameko services, will host and run them. By default this will try to
    find classes that look like services (anything with nameko entrypoints),
    but a specific service can be specified via
    ``nameko run module:ServiceClass``.
    """

    name = 'run'

    @staticmethod
    def init_parser(parser):
        parser.add_argument(
            'services', nargs='+',
            metavar='module[:service class]',
            help='python path to one or more service classes to run')

        parser.add_argument(
            '--broker', default=None,
            help=(
                'Deprecated option for setting up RabbitMQ broker URI. '
                'Use --define or --config and set AMQP_URI instead.'
            )
        )

        parser.add_argument(
            '--backdoor-port', type=int,
            help='Specify a port number to host a backdoor, which can be'
            ' connected to for an interactive interpreter within the running'
            ' service process using `nameko backdoor`.')

        return parser

    @staticmethod
    def main(args):
        from .run import main
        main(args)


class Shell(Command):
    """Launch an interactive python shell for working with remote nameko
    services.

    This is a regular interactive interpreter, with a special module ``n``
    added to the built-in namespace, providing ``n.rpc`` and
    ``n.dispatch_event``.
    """

    name = 'shell'

    SHELLS = ['bpython', 'ipython', 'plain']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--broker', default=None,
            help=(
                'Deprecated option for setting up RabbitMQ broker URI. '
                'Use --define or --config and set up AMQP_URI instead.'
            )
        )
        parser.add_argument(
            '--interface', choices=cls.SHELLS,
            help='Specify an interactive interpreter interface.')
        return parser

    @staticmethod
    def main(args):
        from .shell import main
        main(args)


commands = Command.__subclasses__()  # pylint: disable=E1101



from __future__ import print_function

import os
from subprocess import call

from nameko.exceptions import CommandError


def main(args):
    for choice in ['netcat', 'nc', 'telnet']:
        if os.system('which %s &> /dev/null' % choice) == 0:
            prog = choice
            break
    else:
        raise CommandError('Could not find an installed telnet.')

    target = args.target
    if ':' in target:
        host, port = target.split(':', 1)
    else:
        host, port = 'localhost', target

    rlwrap = args.rlwrap

    cmd = [prog, str(host), str(port)]
    if prog == 'netcat':
        cmd.append('--close')
    if rlwrap is None:
        rlwrap = os.system('which rlwrap &> /dev/null') == 0
    if rlwrap:
        cmd.insert(0, 'rlwrap')
    try:
        if call(cmd) != 0:
            raise CommandError(
                'Backdoor unreachable on {}'.format(target)
            )
    except (EOFError, KeyboardInterrupt):
        print()
        if choice == 'telnet' and rlwrap:
            call(['reset'])



import eventlet
from eventlet import Timeout
from mock import create_autospec

from nameko.containers import ServiceContainer
from nameko.testing.utils import get_extension, wait_for_call
from nameko.timer import Timer, timer


def spawn_managed_thread(fn, identifier=None):
    return eventlet.spawn(fn)


def test_provider():
    container = create_autospec(ServiceContainer)
    container.service_name = "service"
    container.spawn_managed_thread = spawn_managed_thread

    timer = Timer(interval=0.1).bind(container, "method")
    timer.setup()
    timer.start()

    assert timer.interval == 0.1

    with wait_for_call(1, container.spawn_worker) as spawn_worker:
        with Timeout(1):
            timer.stop()

    # the timer should have stopped and should only have spawned
    # a single worker
    spawn_worker.assert_called_once_with(timer, (), {})
    assert timer.gt.dead


def test_stop_timer_immediately():
    container = create_autospec(ServiceContainer)
    container.service_name = "service"
    container.config = {}

    timer = Timer(interval=5).bind(container, "method")
    timer.setup()
    timer.start()

    eventlet.sleep(0.1)
    timer.stop()

    assert container.spawn_worker.call_count == 0
    assert timer.gt.dead


def test_kill_stops_timer():
    container = create_autospec(ServiceContainer)
    container.service_name = "service"
    container.spawn_managed_thread = spawn_managed_thread

    timer = Timer(interval=0).bind(container, "method")
    timer.setup()
    timer.start()

    with wait_for_call(1, container.spawn_worker):
        timer.kill()

    # unless the timer is dead, the following nap would cause a timer
    # to trigger
    eventlet.sleep(0.1)
    assert container.spawn_worker.call_count == 1


class TestEntrypointArguments:

    def test_expected_exceptions_and_sensitive_arguments(self, container_factory):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @timer(1, expected_exceptions=Boom, sensitive_arguments=["arg"])
            def method(self, arg):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, Timer)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["arg"]



from copy import deepcopy


config = {}
"""
It is possible to set the config by using ``--config`` switch of Nameko
command line interface.

The config can be used straight on service definition level, e.g.::

    from nameko import config
    from nameko.messaging import Consumer, Queue

    class Service:

        @consume(
            queue=Queue(
                exchange=config.MY_EXCHANGE,
                routing_key=config.MY_ROUTING_KEY,
                name=config.MY_QUEUE_NAME
            ),
            prefetch_count=config.MY_CONSUMER_PREFETCH_COUNT
        )
        def consume(self, payload):
            pass

As a plain dictionary it is normally mutable at any point. After all,
we're all consenting adults here.)
"""


class setup_config:
    """
    A context manager for setting up a complete config for the given context

    Intended to be used mainly in tests::

        @pytest.yield_fixture
        def config():
            with setup_config({"spam": "ham"}):
                yield

    """

    def __init__(self, context_config):
        self.original_data = deepcopy(config)
        config.clear()
        config.update(context_config)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        config.clear()
        config.update(self.original_data)


class update_config(setup_config):
    """
    A context manager for updating config just for the given context

    Intended to be used mainly in tests:

        @pytest.yield_fixture
        def memory_rabbit_config():
            with update_config({"AMQP_URI": "memory://"}):
                yield
    """

    def __init__(self, context_config):
        self.original_data = deepcopy(config)
        config.update(context_config)



from __future__ import print_function

import argparse
import os
import re
import warnings
from functools import partial

import yaml

from nameko import config
from nameko.constants import AMQP_URI_CONFIG_KEY
from nameko.exceptions import CommandError, ConfigurationError

from . import commands


try:
    import regex
except ImportError:  # pragma: no cover
    ENV_VAR_MATCHER = re.compile(
        r"""
            \$\{       # match characters `${` literally
            ([^}:\s]+) # 1st group: matches any character except `}` or `:`
            :?         # matches the literal `:` character zero or one times
            ([^}]+)?   # 2nd group: matches any character except `}`
            \}         # match character `}` literally
        """, re.VERBOSE
    )
else:  # pragma: no cover
    ENV_VAR_MATCHER = regex.compile(
        r"""
        \$\{                #  match ${
        (                   #  first capturing group: variable name
            [^{}:\s]+       #  variable name without {,},: or spaces
        )
        (?:                 # non capturing optional group for value
            :               # match :
            (               # 2nd capturing group: default value
                (?:         # non capturing group for OR
                    [^{}]   # any non bracket
                |           # OR
                    \{      # literal {
                    (?2)    # recursive 2nd capturing group aka ([^{}]|{(?2)})
                    \}      # literal }
                )*          #
            )
        )?
        \}                  # end of macher }
        """,
        regex.VERBOSE
    )

IMPLICIT_ENV_VAR_MATCHER = re.compile(
    r"""
        .*          # matches any number of any characters
        \$\{.*\}    # matches any number of any characters
                    # between `${` and `}` literally
        .*          # matches any number of any characters
    """, re.VERBOSE
)


def setup_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    for command in commands.commands:
        command_parser = subparsers.add_parser(
            command.name, description=command.__doc__,
            conflict_handler='resolve')
        command.init_parser(command_parser)
        command_parser.set_defaults(main=command.main)
        command_parser.add_argument(
            '--config',
            help='The YAML configuration file')
        command_parser.add_argument(
            '-d', '--define',
            type=parse_config_option, action='append', metavar='KEY=VALUE',
            help='Set config entry. Overrides value loaded from config file.'
            ' Can be used multiple times.'
            ' Example: --define AMQP_URI=pyamqp://guest:guest@localhost')

    return parser


def _replace_env_var(match):
    env_var, default = match.groups()
    value = os.environ.get(env_var, None)
    if value is None:
        # expand default using other vars
        if default is None:
            # regex module return None instead of
            #  '' if engine didn't entered default capture group
            default = ''

        value = default
        while IMPLICIT_ENV_VAR_MATCHER.match(value):  # pragma: no cover
            value = ENV_VAR_MATCHER.sub(_replace_env_var, value)
    return value


def env_var_constructor(loader, node, raw=False):
    raw_value = loader.construct_scalar(node)
    value = ENV_VAR_MATCHER.sub(_replace_env_var, raw_value)
    return value if raw else yaml.safe_load(value)


def setup_yaml_parser():
    yaml.add_constructor('!env_var', env_var_constructor)
    yaml.add_constructor('!raw_env_var',
                         partial(env_var_constructor, raw=True))
    yaml.add_implicit_resolver('!env_var', IMPLICIT_ENV_VAR_MATCHER)


def load_config(config_path):
    with open(config_path) as fle:
        return yaml.load(fle)


def parse_config_option(text):
    if '=' in text:
        key, value = text.strip().split('=', 1)
        return key, yaml.load(value)
    else:
        return text, True


def setup_config(args):
    setup_yaml_parser()
    if args.config:
        config.update(load_config(args.config))
    if hasattr(args, "broker") and args.broker:
        warnings.warn(
            (
                "--broker option is going to be removed. ",
                "Use --config or --define and set AMQP_URI instead."
            ),
            DeprecationWarning
        )
        if AMQP_URI_CONFIG_KEY not in config:
            config.update({AMQP_URI_CONFIG_KEY: args.broker})
    if args.define:
        config.update(args.define)


def main():
    parser = setup_parser()
    args = parser.parse_args()
    setup_config(args)
    try:
        args.main(args)
    except (CommandError, ConfigurationError) as exc:
        print("Error: {}".format(exc))



import eventlet
import pytest
from mock import DEFAULT, patch

from nameko.cli.commands import Backdoor
from nameko.cli.main import setup_parser
from nameko.cli.run import setup_backdoor
from nameko.exceptions import CommandError


@pytest.fixture
def running_backdoor():
    runner = object()
    green_socket, gt = setup_backdoor(runner, 0)
    eventlet.sleep(0)  # give backdoor a chance to spawn
    socket_name = green_socket.fd.getsockname()
    return socket_name


def test_no_telnet():
    parser = setup_parser()
    args = parser.parse_args(['backdoor', '0'])
    with patch('nameko.cli.backdoor.os') as mock_os:
        mock_os.system.return_value = -1
        with pytest.raises(CommandError) as exc:
            Backdoor.main(args)
    assert 'Could not find an installed telnet' in str(exc)


def test_no_running_backdoor():
    parser = setup_parser()
    args = parser.parse_args(['backdoor', '0'])
    with patch.multiple(
        'nameko.cli.backdoor', call=DEFAULT, os=DEFAULT
    ) as mocks:
        mocks['os'].system.return_value = 0
        mocks['call'].return_value = -1
        with pytest.raises(CommandError) as exc:
            Backdoor.main(args)
    assert 'Backdoor unreachable' in str(exc)


def test_basic(running_backdoor):
    socket_arg = '{}:{}'.format(*running_backdoor)
    parser = setup_parser()
    args = parser.parse_args(['backdoor', socket_arg])
    with patch.multiple(
        'nameko.cli.backdoor', call=DEFAULT, os=DEFAULT
    ) as mocks:
        mock_call = mocks['call']
        mocks['os'].system.return_value = 0
        mock_call.return_value = 0
        Backdoor.main(args)
    (cmd, ), _ = mock_call.call_args
    expected = (
        ['rlwrap', 'netcat'] +
        list(map(str, running_backdoor)) +
        ['--close']
    )
    assert cmd == expected


def test_default_host(running_backdoor):
    _, port = running_backdoor
    parser = setup_parser()
    args = parser.parse_args(['backdoor', str(port)])
    with patch.multiple(
        'nameko.cli.backdoor', call=DEFAULT, os=DEFAULT
    ) as mocks:
        mock_call = mocks['call']
        mocks['os'].system.return_value = 0
        mock_call.return_value = 0
        Backdoor.main(args)
    (cmd, ), _ = mock_call.call_args
    expected = ['rlwrap', 'netcat', 'localhost'] + [str(port)] + ['--close']
    assert cmd == expected


def test_stop(running_backdoor):
    _, port = running_backdoor
    parser = setup_parser()
    args = parser.parse_args(['backdoor', str(port)])
    with patch.multiple(
        'nameko.cli.backdoor', call=DEFAULT, os=DEFAULT
    ) as mocks:
        # choose telnet (skip nc and netcat) and find rlwrap
        mocks['os'].system.side_effect = [-1, -1, 0, 0]
        mocks['call'].side_effect = [KeyboardInterrupt, 0]
        Backdoor.main(args)



import pytest

from nameko import config, setup_config, update_config


@pytest.mark.usefixtures("empty_config")
def test_setup_config():

    assert config == {}

    setup_config({"HEARTBEAT": 10})

    assert config == {"HEARTBEAT": 10}

    setup_config({})

    assert config == {}


@pytest.mark.usefixtures("empty_config")
def test_setup_config_context_manager():

    with setup_config({"HEARTBEAT": 10}):
        assert config == {"HEARTBEAT": 10}

    assert config == {}

    with setup_config({"HEARTBEAT": 40}):
        assert config == {"HEARTBEAT": 40}
        with setup_config({"HEARTBEAT": 10}):
            assert config == {"HEARTBEAT": 10}
        assert config == {"HEARTBEAT": 40}

    assert config == {}


@pytest.mark.usefixtures("empty_config")
def test_setup_config_replaces_whole_config():

    setup_config({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    with setup_config({"HEARTBEAT": 40}):
        assert config == {"HEARTBEAT": 40}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}


@pytest.mark.usefixtures("empty_config")
def test_update_config():

    setup_config({"HEARTBEAT": 10, "ACCEPT": "YAML"})

    with update_config({"HEARTBEAT": 40}):
        assert config == {"HEARTBEAT": 40, "ACCEPT": "YAML"}

    assert config == {"HEARTBEAT": 10, "ACCEPT": "YAML"}



from __future__ import absolute_import

import time
from logging import getLogger

from eventlet import Timeout
from eventlet.event import Event

from nameko.extensions import Entrypoint


_log = getLogger(__name__)


class Timer(Entrypoint):
    def __init__(self, interval, **kwargs):
        """
        Timer entrypoint implementation. Fires every :attr:`self.interval`
        seconds.

        The implementation sleeps first, i.e. does not fire at time 0.

        Example::

            timer = Timer.decorator

            class Service(object):
                name = "service"

                @timer(interval=5)
                def tick(self):
                    pass

        """
        self.interval = interval
        self.should_stop = Event()
        self.gt = None
        super(Timer, self).__init__(**kwargs)

    def start(self):
        _log.debug('starting %s', self)
        self.gt = self.container.spawn_managed_thread(self._run)

    def stop(self):
        _log.debug('stopping %s', self)
        self.should_stop.send(True)
        self.gt.wait()

    def kill(self):
        _log.debug('killing %s', self)
        self.gt.kill()

    def _run(self):
        """ Runs the interval loop. """

        sleep_time = self.interval

        while True:
            # sleep for `sleep_time`, unless `should_stop` fires, in which
            # case we leave the while loop and stop entirely
            with Timeout(sleep_time, exception=False):
                self.should_stop.wait()
                break

            start = time.time()

            self.handle_timer_tick()

            elapsed_time = (time.time() - start)

            # next time, sleep however long is left of our interval, taking
            # off the time we took to run
            sleep_time = max(self.interval - elapsed_time, 0)

    def handle_timer_tick(self):
        args = ()
        kwargs = {}

        # Note that we don't catch ContainerBeingKilled here. If that's raised,
        # there is nothing for us to do anyway. The exception bubbles, and is
        # caught by :meth:`Container._handle_thread_exited`, though the
        # triggered `kill` is a no-op, since the container is alredy
        # `_being_killed`.
        self.container.spawn_worker(self, args, kwargs)


timer = Timer.decorator



import sys

import pytest
from mock import patch

from nameko.cli.main import main


@pytest.fixture
def command():

    def _command(*argv):
        with patch.object(sys, 'argv', argv):
            main()

    return _command



from nameko import setup_config
from nameko.standalone.events import event_dispatcher

setup_config({
    'AMQP_URI': AMQP_URI  # e.g. "pyamqp://guest:guest@localhost"
})

dispatch = event_dispatcher()
dispatch("service_a", "event_type", "payløad")



from textwrap import dedent

from mock import patch


@patch('nameko.cli.main.os')
def test_main(mock_os, tmpdir, capsys, command):

    config_file = tmpdir.join('config.yaml')
    config_file.write("""
        FOO: ${FOO:foobar}
        BAR: ${BAR}
    """)

    mock_os.environ = {
        'BAR': '[1,2,3]'
    }

    command(
        'nameko', 'show-config',
        '--config', config_file.strpath,
    )

    out, _ = capsys.readouterr()

    expected = dedent("""
        BAR:
        - 1
        - 2
        - 3
        FOO: foobar
    """).strip()

    assert out.strip() == expected



import os
import sys

import pytest
import yaml
from mock import patch

from nameko.cli.main import (
    ENV_VAR_MATCHER, main, parse_config_option, setup_parser,
    setup_yaml_parser
)
from nameko.exceptions import CommandError, ConfigurationError


try:
    import regex
except ImportError:  # pragma: no cover
    has_regex_module = False
else:  # pragma: no cover
    del regex
    has_regex_module = True


@pytest.yield_fixture(autouse=True)
def fake_argv(empty_config):
    with patch.object(
        sys,
        'argv',
        [
            'nameko',
            'run',
            '--define',
            'AMQP_URI=pyamqp://someuser:*****@somehost/',
            'test.sample:Service',
        ],
    ):
        yield


def test_run():
    with patch('nameko.cli.run.main') as run:
        main()
    assert run.call_count == 1
    (args,), _ = run.call_args
    assert args.define == [('AMQP_URI', 'pyamqp://someuser:*****@somehost/')]


@pytest.mark.parametrize('exception', (CommandError, ConfigurationError))
def test_error(exception, capsys):
    with patch('nameko.cli.run.main') as run:
        run.side_effect = exception('boom')
        main()
    out, _ = capsys.readouterr()
    assert out.strip() == 'Error: boom'


@pytest.mark.parametrize(('param', 'value'), (
    (None, None),
    ('--rlwrap', True),
    ('--no-rlwrap', False),
))
def test_flag_action(param, value):
    parser = setup_parser()
    args = ['backdoor', 0]
    if param is not None:
        args.append(param)
    parsed = parser.parse_args(args)
    assert parsed.rlwrap is value


@pytest.mark.parametrize(
    ('text', 'expected_key', 'expected_value'),
    (
        # strings
        ('SPAM=ham', 'SPAM', 'ham'),
        ('  SPAM=ham  ', 'SPAM', 'ham'),
        ('SPAM="ham"', 'SPAM', 'ham'),
        ("SPAM='ham'", 'SPAM', 'ham'),
        ('SPAM="1"', 'SPAM', "1"),
        # simple types
        ('SPAM=1', 'SPAM', 1),
        ('SPAM=1.0', 'SPAM', 1.0),
        ('SPAM=True', 'SPAM', True),
        ('SPAM=False', 'SPAM', False),
        ('SPAM', 'SPAM', True),
        ('SPAM=', 'SPAM', None),
        # structure
        ('SPAM=[{"egg": "ham"}]', 'SPAM', [{'egg': 'ham'}]),
        # deals with equal signs in value
        ('SPAM=EGG=HAM', 'SPAM', 'EGG=HAM'),
    )
)
def test_parse_config_option(text, expected_key, expected_value):
    assert parse_config_option(text) == (expected_key, expected_value)


class TestConfigEnvironmentParser(object):

    @pytest.mark.parametrize(('value', 'expected'), [
        # no var
        ('raw', []),
        # var without default
        ('${VAR_NAME}', [('VAR_NAME', '')]),
        # on var with default value
        ('${VAR_NAME:default}', [('VAR_NAME', 'default')]),
        # multiple match with default
        ('${VAR_NAME1:default1}_${VAR_NAME2:default2}', [
            ('VAR_NAME1', 'default1'),
            ('VAR_NAME2', 'default2'),
        ]),
    ])
    def test_maching_env_variable(self, value, expected):
        res = ENV_VAR_MATCHER.findall(value)
        assert res == expected

    @pytest.mark.parametrize(('value', 'expected'), [
        # on var, with {} in default (as str.format accept)
        ('${VAR_NAME:my-{pytemplate}-template}', [
            ('VAR_NAME', 'my-{pytemplate}-template')]),
        # var with var in default
        ('${VAR_NAME:my-${OTHER_VAR}-template}', [
            ('VAR_NAME', 'my-${OTHER_VAR}-template')]),
        # recursive interpretation of last var
        ('my-${OTHER_VAR}-template', [
            ('OTHER_VAR', '')]),
        # var with var in default in default
        ('${VAR_NAME:my-${OTHER_VAR:1}-template}', [
            ('VAR_NAME', 'my-${OTHER_VAR:1}-template')]),
        # recursive interpretation of last var
        ('my-${OTHER_VAR:1}-template', [
            ('OTHER_VAR', '1')]),
        # multiple var in default
        ('${VAR_NAME1:default1}_${VAR_NAME2:default2}', [
            ('VAR_NAME1', 'default1'),
            ('VAR_NAME2', 'default2'),
        ]),
    ])
    @pytest.mark.skipif(not has_regex_module,
                        reason='0 support for nested env without regex module')
    def test_maching_recursive_with_regex(
            self, value, expected
    ):  # pragma: no cover
        res = ENV_VAR_MATCHER.findall(value)
        assert res == expected


class TestConfigEnvironmentVariables(object):

    @pytest.mark.parametrize(('yaml_config', 'env_vars', 'expected_config'), [
        # no default value, no env value
        ('FOO: ${BAR}', {}, {'FOO': None}),
        # use default value if env value not provided
        ('FOO: ${BAR:foo}', {}, {'FOO': 'foo'}),
        # use env value
        ('FOO: ${BAR}', {'BAR': 'bar'}, {'FOO': 'bar'}),
        # use env value instead of default
        ('FOO: ${BAR:foo}', {'BAR': 'bar'}, {'FOO': 'bar'}),
        # handles multi line
        (
            """
            FOO: ${BAR:foo}
            BAR: ${FOO:bar}
            """,
            {'BAR': 'bar', 'FOO': 'foo'},
            {'FOO': 'bar', 'BAR': 'foo'}
        ),
        # quoted values don't work without explicit resolver
        ('FOO: "${BAR:foo}"', {'BAR': 'bar'}, {'FOO': '${BAR:foo}'}),
        # quoted values work only with explicit resolver
        ('FOO: !env_var "${BAR:foo}"', {'BAR': 'bar'}, {'FOO': 'bar'}),
        # quoted values with raw_env_var constructor to avoid yaml parsing
        ('FOO: !raw_env_var "${BAR}"', {'BAR': '123.0'}, {'FOO': '123.0'}),
        # $ sign can be used
        ('FOO: $bar', {}, {'FOO': '$bar'}),
        # multiple substitutions
        (
            'FOO: http://${BAR:foo}/${FOO:bar}',
            {'BAR': 'bar', 'FOO': 'foo'},
            {'FOO': 'http://bar/foo'}
        ),
        # can handle int, float and boolean
        (
            """
            FOO:
                - BAR: ${INT:1}
                - BAR: ${FLOAT:1.1}
                - BAR: ${BOOL:True}
            """,
            {}, {'FOO': [{'BAR': 1}, {'BAR': 1.1}, {'BAR': True}]}
        ),
        (
            """
            FOO:
                - BAR: ${INT}
                - BAR: ${FLOAT}
                - BAR: ${BOOL}
            """,
            {'INT': '1', 'FLOAT': '1.1', 'BOOL': 'True'},
            {'FOO': [{'BAR': 1}, {'BAR': 1.1}, {'BAR': True}]}
        ),
        # quoted default results in string
        (
            """
            FOO:
                - BAR: ${INT:"1"}
                - BAR: ${FLOAT:"1.1"}
                - BAR: ${BOOL:"True"}
            """,
            {}, {'FOO': [{'BAR': "1"}, {'BAR': "1.1"}, {'BAR': "True"}]}
        ),
        # quoted value results in string
        # double-quote required
        (
            """
            FOO:
                - BAR: !env_var "'${INT}'"
                - BAR: !env_var "'${FLOAT}'"
                - BAR: !env_var "'${BOOL}'"
            """,
            {'INT': '1', 'FLOAT': '1.1', 'BOOL': 'True'},
            {'FOO': [{'BAR': "1"}, {'BAR': "1.1"}, {'BAR': "True"}]}
        ),
        # quoted input results in string
        # double-quote required
        (
            """
            FOO:
                - BAR: ${INT}
                - BAR: ${FLOAT}
                - BAR: ${BOOL}
            """,
            {'INT': '"1"', 'FLOAT': '"1.1"', 'BOOL': '"True"'},
            {'FOO': [{'BAR': "1"}, {'BAR': "1.1"}, {'BAR': "True"}]}
        ),
        # list of scalar values
        (
            """
            FOO:
                - ${INT_1}
                - ${INT_2}
                - ${INT_3}
            BAR:
                - 1
                - 2
                - 3
            """,
            {'INT_1': '1', 'INT_2': '2', 'INT_3': '3'},
            {'FOO': [1, 2, 3], 'BAR': [1, 2, 3]}
        ),
        # inline list
        (
            """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
            {"LIST": "[1,2,3]"},
            {'FOO': [1, 2, 3], 'BAR': [1, 2, 3]}
        ),
        # inline list with block style
        (
            """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
            {"LIST": "- 1\n- 2\n- 3"},
            {'FOO': [1, 2, 3], 'BAR': [1, 2, 3]}
        ),
        # inline list via explicit tag
        (
            """
            FOO: !env_var "${LIST}"
            BAR:
                - 1
                - 2
                - 3
            """,
            {"LIST": "[1,2,3]"},
            {'FOO': [1, 2, 3], 'BAR': [1, 2, 3]}
        ),
        # inline list with default
        (
            """
            FOO: ${LIST:[1,2,3]}
            BAR:
                - 1
                - 2
                - 3
            """,
            {},
            {'FOO': [1, 2, 3], 'BAR': [1, 2, 3]}
        ),
        # inline list containing list
        (
            """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
            {"LIST": "[1, 2, 3, ['a', 'b', 'c']]"},
            {'FOO': [1, 2, 3, ['a', 'b', 'c']], 'BAR': [1, 2, 3]}
        ),
        # inline dict
        (
            """
            FOO: ${DICT}
            BAR:
                - 1
                - 2
                - 3
            """,
            {"DICT": "{'one': 1, 'two': 2}"},
            {'FOO': {'one': 1, 'two': 2}, 'BAR': [1, 2, 3]}
        ),
        # inline dict with block style
        (
            """
            FOO: ${DICT}
            BAR:
                - 1
                - 2
                - 3
            """,
            {"DICT": "one: 1\ntwo: 2"},
            {'FOO': {'one': 1, 'two': 2}, 'BAR': [1, 2, 3]}
        ),
    ])
    def test_environment_vars_in_config(
        self, yaml_config, env_vars, expected_config
    ):
        setup_yaml_parser()

        with patch.dict('os.environ'):
            for key, val in env_vars.items():
                os.environ[key] = val

            results = yaml.load(yaml_config)
            assert results == expected_config

    def test_cannot_recurse(self):

        setup_yaml_parser()

        yaml_config = """
            FOO: ${VAR1}
            BAR:
                - 1
                - 2
                - 3
        """

        with patch.dict('os.environ'):
            os.environ["VAR1"] = "${VAR1}"

            results = yaml.load(yaml_config)
            assert results == {'FOO': "${VAR1}", 'BAR': [1, 2, 3]}

    @pytest.mark.parametrize(('yaml_config', 'env_vars', 'expected_config'), [
        # recursive env with root value
        (
            """
            FOO: ${FOO:val_${INDICE:1}}
            """,
            {"FOO": 'val_a'},
            {"FOO": 'val_a'},
        ),
        # recursive env with root default and sub value
        (
            """
            FOO: ${FOO:val_${INDICE:1}}
            """,
            {"INDICE": 'b'},
            {"FOO": 'val_b'},
        ),
        # recursive env requiring data
        (
            """
            FOO: ${FOO:val_${INDICE}}
            """,
            {"INDICE": 'b'},
            {"FOO": 'val_b'},
        ),

        # default data with bracket
        (
            """
            FOO: ${FOO:"{name} {age}"}
            """,
            {},
            {"FOO": '{name} {age}'},
        ),
        # default data with bracket
        (
            """
            FOO: ${FOO:"{name} {age}"}
            """,
            {"FOO": '"{surname}"'},
            {"FOO": '{surname}'},
        ),
    ])
    @pytest.mark.skipif(not has_regex_module,
                        reason='0 support for nested env without regex module')
    def test_environment_vars_recursive_in_config(
            self, yaml_config, env_vars, expected_config
    ):  # pragma: no cover
        setup_yaml_parser()

        with patch.dict('os.environ'):
            for key, val in env_vars.items():
                os.environ[key] = val

            results = yaml.load(yaml_config)
            assert results == expected_config

    @pytest.mark.parametrize(('yaml_config', 'env_vars', 'expected_config'), [
        # recursive env with root value
        (
            """
            FOO: ${FOO:val_${INDICE:1}}
            """,
            {"FOO": 'val_a'},
            {"FOO": 'val_a}'},
        ),
        # recursive env with root default and sub value
        (
            """
            FOO: ${FOO:val_${INDICE:1}}
            """,
            {"INDICE": 'b'},
            {"FOO": 'val_${INDICE:1}'},
        ),
        # recursive env requiring data
        (
            """
            FOO: ${FOO:${INDICE}_val}
            """,
            {"INDICE": 'b'},
            {"FOO": '${INDICE_val}'},
        ),

    ])
    @pytest.mark.skipif(has_regex_module,
                        reason='default behavior if no regex module')
    def test_unhandled_recursion(
            self, yaml_config, env_vars, expected_config
    ):  # pragma: no cover
        setup_yaml_parser()

        with patch.dict('os.environ'):
            for key, val in env_vars.items():
                os.environ[key] = val
            results = yaml.load(yaml_config)
            assert results == expected_config



from nameko import setup_config
from nameko.standalone.rpc import ClusterRpcClient


setup_config({
    'AMQP_URI': AMQP_URI  # e.g. "pyamqp://guest:guest@localhost"
})

with ClusterRpcClient() as cluster_rpc:
    cluster_rpc.service_x.remote_method("hellø")  # "hellø-x-y"



import os
import sys

import pytest
from mock import Mock, patch

from nameko.cli.shell import make_nameko_helper
from nameko.constants import (
    AMQP_URI_CONFIG_KEY, SERIALIZER_CONFIG_KEY, WEB_SERVER_CONFIG_KEY
)
from nameko.rpc import Client


def test_helper_module(rabbit_config):
    helper = make_nameko_helper()
    assert isinstance(helper.rpc, Client)
    helper.disconnect()


@pytest.yield_fixture
def pystartup(tmpdir):
    startup = tmpdir.join('startup.py')
    startup.write('foo = 42')

    with patch.dict(os.environ, {'PYTHONSTARTUP': str(startup)}):
        yield


@pytest.yield_fixture(autouse=True)
def fake_alternative_interpreters():
    # Make sure these appear unavailable even if installed. We cheat slightly,
    # and have the call to `embed` raise the ImportError, rather than the
    # actual module import (this is easier to do and has the same effect in our
    # case).
    fake_module = Mock()
    fake_module.embed.side_effect = ImportError
    with patch.dict(sys.modules, {
        'IPython': fake_module,
        'bpython': fake_module,
    }):
        yield


@pytest.mark.usefixtures('rabbit_config')
def test_basic(command, pystartup):
    with patch('nameko.cli.shell.code') as code:
        command('nameko', 'shell')

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['foo'] == 42
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_plain(command, pystartup):
    with patch('nameko.cli.shell.code') as code:
        command(
            'nameko', 'shell',
            '--interface', 'plain'
        )

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['foo'] == 42
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_plain_fallback(command, pystartup):
    with patch('nameko.cli.shell.code') as code:
        command(
            'nameko', 'shell',
            '--interface', 'bpython'
        )

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['foo'] == 42
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_bpython(command, pystartup):
    with patch('bpython.embed') as embed:
        command(
            'nameko', 'shell',
            '--interface', 'bpython'
        )

    _, kwargs = embed.call_args
    local = kwargs['locals_']
    assert 'n' in local.keys()
    assert local['foo'] == 42
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_ipython(command, pystartup):
    with patch('IPython.embed') as embed:
        command(
            'nameko', 'shell',
            '--interface', 'ipython'
        )

    _, kwargs = embed.call_args
    local = kwargs['user_ns']
    assert 'n' in local.keys()
    assert local['foo'] == 42
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_config(command, pystartup, tmpdir):

    config_file = tmpdir.join('config.yaml')
    config_file.write("""
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """)

    with patch('nameko.cli.shell.code') as code:
        command(
            'nameko', 'shell',
            '--config', config_file.strpath,
        )

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['n'].config[WEB_SERVER_CONFIG_KEY] == '0.0.0.0:8001'
    assert local['n'].config[SERIALIZER_CONFIG_KEY] == 'json'
    local['n'].disconnect()


@pytest.mark.usefixtures('rabbit_config')
def test_config_options(command, pystartup, tmpdir):

    config_file = tmpdir.join('config.yaml')
    config_file.write("""
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """)

    with patch('nameko.cli.shell.code') as code:
        command(
            'nameko', 'shell',
            '--config', config_file.strpath,
            '--define', 'serializer=pickle',
            '--define', 'EGG=[{"spam": True}]',
        )

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['n'].config[WEB_SERVER_CONFIG_KEY] == '0.0.0.0:8001'
    assert local['n'].config[SERIALIZER_CONFIG_KEY] == 'pickle'
    assert local['n'].config['EGG'] == [{'spam': True}]
    local['n'].disconnect()


@pytest.mark.usefixtures('empty_config')
class TestBanner(object):

    @pytest.yield_fixture(autouse=True)
    def patch_nameko_helper(self):
        with patch('nameko.cli.shell.make_nameko_helper'):
            yield

    def test_banner(self, command, tmpdir):

        amqp_uri = "amqp://broker/config"

        config_file = tmpdir.join('config.yaml')
        config_file.write("""
            WEB_SERVER_ADDRESS: '0.0.0.0:8001'
            AMQP_URI: '{}'
            serializer: 'json'
        """.format(amqp_uri))

        with patch('nameko.cli.shell.ShellRunner') as shell_runner:
            command(
                'nameko', 'shell',
                '--config', config_file.strpath
            )

        expected_message = (
            "Broker: {}".format(amqp_uri)
        )
        (banner, _), _ = shell_runner.call_args
        assert expected_message in banner


@pytest.mark.usefixtures('empty_config')
def test_broker_option_deprecated(command, rabbit_uri):

    with patch('nameko.cli.shell.code') as code:
        with patch('nameko.cli.main.warnings') as warnings:
            command(
                'nameko', 'shell',
                '--broker', rabbit_uri,
            )

    _, kwargs = code.interact.call_args
    local = kwargs['local']
    assert 'n' in local.keys()
    assert local['n'].config[AMQP_URI_CONFIG_KEY] == rabbit_uri
    local['n'].disconnect()

    assert warnings.warn.call_count



import inspect
import re
import sys
from copy import deepcopy
from pydoc import locate

import eventlet
import six
from eventlet.queue import LightQueue

REDACTED = "********"


def get_redacted_args(entrypoint, *args, **kwargs):
    """ Utility function for use with entrypoints that are marked with
    ``sensitive_arguments`` -- e.g. :class:`nameko.rpc.Rpc` and
    :class:`nameko.events.EventHandler`.

    :Parameters:
        entrypoint : :class:`~nameko.extensions.Entrypoint`
            The entrypoint that fired.
        args : tuple
            Positional arguments for the method call.
        kwargs : dict
            Keyword arguments for the method call.

    The entrypoint should have a ``sensitive_arguments`` attribute, the value
    of which is a string or tuple of strings specifying the arguments or
    partial arguments that should be redacted. To partially redact an argument,
    the following syntax is used::

        <argument-name>.<dict-key>[<list-index>]

    :Returns:
        A dictionary as returned by :func:`inspect.getcallargs`, but with
        sensitive arguments or partial arguments redacted.

    .. note::

        This function does not raise if one of the ``sensitive_arguments``
        doesn't match or partially match the calling ``args`` and ``kwargs``.
        This allows "fuzzier" pattern matching (e.g. redact a field if it is
        present, and otherwise do nothing).

        To avoid exposing sensitive arguments through a typo, it is recommend
        to test the configuration of each entrypoint with
        ``sensitive_arguments`` individually. For example:

        .. code-block:: python

            class Service(object):
                @rpc(sensitive_arguments="foo.bar")
                def method(self, foo):
                    pass

            container = ServiceContainer(Service, {})
            entrypoint = get_extension(container, Rpc, method_name="method")

            # no redaction
            foo = "arg"
            expected_foo = {'foo': "arg"}
            assert get_redacted_args(entrypoint, foo) == expected

            # 'bar' key redacted
            foo = {'bar': "secret value", 'baz': "normal value"}
            expected = {'foo': {'bar': "********", 'baz': "normal value"}}
            assert get_redacted_args(entrypoint, foo) == expected

    .. seealso::

        The tests for this utility demonstrate its full usage:
        :class:`test.test_utils.TestGetRedactedArgs`

    """
    sensitive_arguments = entrypoint.sensitive_arguments
    if isinstance(sensitive_arguments, six.string_types):
        sensitive_arguments = (sensitive_arguments,)

    method = getattr(entrypoint.container.service_cls, entrypoint.method_name)
    callargs = inspect.getcallargs(method, None, *args, **kwargs)
    del callargs['self']

    # make a deepcopy before redacting so that "partial" redacations aren't
    # applied to a referenced object
    callargs = deepcopy(callargs)

    def redact(data, keys):
        key = keys[0]
        if len(keys) == 1:
            try:
                data[key] = REDACTED
            except (KeyError, IndexError, TypeError):
                pass
        else:
            if key in data:
                redact(data[key], keys[1:])

    for variable in sensitive_arguments:
        keys = []
        for dict_key, list_index in re.findall(r"(\w+)|\[(\d+)\]", variable):
            if dict_key:
                keys.append(dict_key)
            elif list_index:
                keys.append(int(list_index))

        if keys[0] in callargs:
            redact(callargs, keys)

    return callargs


def fail_fast_imap(pool, call, items):
    """ Run a function against each item in a given list, yielding each
    function result in turn, where the function call is handled in a
    :class:`~eventlet.greenthread.GreenThread` spawned by the provided pool.

    If any function raises an exception, all other ongoing threads are killed,
    and the exception is raised to the caller.

    This function is similar to :meth:`~eventlet.greenpool.GreenPool.imap`.

    :param pool: Pool to spawn function threads from
    :type pool: eventlet.greenpool.GreenPool
    :param call: Function call to make, expecting to receive an item from the
        given list
    """
    result_queue = LightQueue(maxsize=len(items))
    spawned_threads = set()

    def handle_result(finished_thread):
        try:
            thread_result = finished_thread.wait()
            spawned_threads.remove(finished_thread)
            result_queue.put((thread_result, None))
        except Exception:
            spawned_threads.remove(finished_thread)
            result_queue.put((None, sys.exc_info()))

    for item in items:
        gt = pool.spawn(call, item)
        spawned_threads.add(gt)
        gt.link(handle_result)

    while spawned_threads:
        result, exc_info = result_queue.get()
        if exc_info is not None:
            # Kill all other ongoing threads
            for ongoing_thread in spawned_threads:
                ongoing_thread.kill()
            # simply raising here (even raising a full exc_info) isn't
            # sufficient to preserve the original stack trace.
            # greenlet.throw() achieves this.
            eventlet.getcurrent().throw(*exc_info)
        yield result


class SpawningProxy(object):
    def __init__(self, items, abort_on_error=False):
        """ Wraps an iterable set of items such that a call on the returned
        SpawningProxy instance will spawn a call in a
        :class:`~eventlet.greenthread.GreenThread` for each item.

        Returns when every spawned thread has completed.

        :param items: Iterable item set to process
        :param abort_on_error: If True, any exceptions raised on an individual
            item call will cause all peer item call threads to be killed, and
            for the exception to be propagated to the caller immediately.
        """
        self._items = items
        self.abort_on_error = abort_on_error

    def __getattr__(self, name):

        def spawning_method(*args, **kwargs):
            items = self._items
            if items:
                pool = eventlet.GreenPool(len(items))

                def call(item):
                    return getattr(item, name)(*args, **kwargs)

                if self.abort_on_error:
                    return list(fail_fast_imap(pool, call, self._items))
                else:
                    return list(pool.imap(call, self._items))
        return spawning_method


class SpawningSet(set):
    """ A set with an ``.all`` property that will spawn a method call on each
    item in the set into its own (parallel) greenthread.
    """
    @property
    def all(self):
        return SpawningProxy(self)


def import_from_path(path):
    """ Import and return the object at `path` if it exists.

    Raises an :exc:`ImportError` if the object is not found.
    """
    if path is None:
        return

    obj = locate(path)
    if obj is None:
        raise ImportError(
            "`{}` could not be imported".format(path)
        )

    return obj
